from datetime import date, timedelta

import jaydebeapi
from pandas import DataFrame
from pkg_resources import resource_filename

from tcegoframework.cfgparsing import (get_dremio_connection,
                                       get_dremio_password, get_dremio_user)

# Variuaveis
# Caminho para driver JDBC do Dremio.
# pathDremioJDBC = 'dremio-jdbc-driver-14.0.0-202103011714040666-9a0c2e10.jar'
path_dremio_driver = resource_filename(
    'tcegoframework.resources', 'dremio-jdbc-driver.jar')

# Consulta Base Dremio - Retorna dados do ultimo dia de recepção dos dados
base_query = 'SELECT "Exercício do orçamento (Ano)" ' \
    ' ,"Órgão (Código/Nome)"' \
    ' ,"Órgão Sucessor Atual (Código/Nome)"' \
    ' ,"Tipo Administração (Nome)"' \
    ' ,"Tipo Poder (Nome)"' \
    ' ,"Classificação orçamentária (Descrição)"' \
    ' ,"Função (Cod/Nome)"' \
    ' ,"Subfunção (Cod/Nome)"' \
    ' ,"Programa (Cod/Nome)"' \
    ' ,"Ação (Cod/Nome)"' \
    ' ,"Natureza Despesa (Cod)"' \
    ' ,"Natureza Despesa (Nome)"' \
    ' ,"Grupo Despesa (Cod/Nome)"' \
    ' ,"Elemento Despesa (Cod/Nome)"' \
    ' ,"Formalidade (Nome)"' \
    ' ,"Modalidade Licitação (Nome)"' \
    ' ,"Fonte Recurso (Cod)"' \
    ' ,"Fonte Recurso (Nome)"' \
    ' ,"Beneficiário (Nome)"' \
    ' ,"Beneficiário (CPF)"' \
    ' ,"Beneficiário (CNPJ)"' \
    ' ,"Beneficiário (CPF/CNPJ)"' \
    ' ,"Período (Dia/Mes/Ano)"' \
    ' ,"Empenho (Sequencial Empenho)"' \
    ' ,"Empenho (Histórico)"' \
    ' ,"Empenho (Número do Processo)"' \
    ' ,"Valor Empenhado"' \
    ' ,"Valor Anulação Empenho"' \
    ' ,"Valor Estorno Anulação Empenho"' \
    ' ,"Valor Cancelamento Empenho"' \
    ' ,"Valor Anulação Cancelamento Empenho"' \
    ' ,"Valor Liquidação Empenho"' \
    ' ,"Valor Anulação Liquidacao Empenho"' \
    ' ,"Valor Ordem de Pagamento"' \
    ' ,"Valor Guia Recolhimento"' \
    ' ,"Valor Anulação Ordem de Pagamento"' \
    ' ,"Valor Estorno Anulação O. Pagamento"' \
    ' ,"Valor Estorno Guia Recolhimento"' \
    ' ,"Valor Saldo do Empenho"' \
    ' ,"Valor Saldo Liquidado"' \
    ' ,"Valor Saldo Pago"' \
    ' ,"Valor Saldo a Pagar"' \
    ' ,"Valor a Liquidar"' \
    ' ,"Valor a Pagar Liquidado" ' \
    ' FROM IE.EXTRACAO.EOFClassificacao c ' \



def get_connection():
    conn = jaydebeapi.connect(
        'com.dremio.jdbc.Driver',
        get_dremio_connection(),
        [
            get_dremio_user(),
            get_dremio_password()
        ],
        path_dremio_driver
    )

    return conn


def execute_query(query):
    conn = get_connection()
    cursor = conn.cursor()
    cursor.execute(query)
    data = DataFrame(
        cursor.fetchall(),
        columns=[i[0] for i in cursor.description]
    )
    conn.close()
    return data


def get_train_data():
    query = base_query + ' WHERE c."Exercício do orçamento (Ano)" >= 2015'
    data = execute_query(query)
    return data


def construct_query(filters):
    # dates daterange orgaos

    conditions = []

    if 'dates' in filters:
        for day in filters['dates']:
            day = day.strftime('%Y/%m/%d')
            condition = f'c."Período (Dia/Mes/Ano)" = \'{day}\''
            conditions.append(condition)
    elif 'daterange' in filters:
        date_a, date_b = filters['daterange']
        if date_a > date_b:
            # p < a and p > b
            date_a = date_a.strftime('%Y/%m/%d')
            condition = f'c."Período (Dia/Mes/Ano)" <= \'{date_a}\''
            conditions.append(condition)
            date_b = date_b.strftime('%Y/%m/%d')
            condition = f'c."Período (Dia/Mes/Ano)" >= \'{date_b}\''
            conditions.append(condition)
        else:
            date_a = date_a.strftime('%Y/%m/%d')
            condition = f'c."Período (Dia/Mes/Ano)" >= \'{date_a}\''
            conditions.append(condition)
            date_b = date_b.strftime('%Y/%m/%d')
            condition = f'c."Período (Dia/Mes/Ano)" <= \'{date_b}\''
            conditions.append(condition)
    else:
        # DIA ANTERIOR
        previous_day = (date.today() - timedelta(days=1)).strftime('%Y/%m/%d')
        condition = f'c."Período (Dia/Mes/Ano)" = \'{previous_day}\''
        conditions.append(condition)
    if 'orgaos' in filters:
        for orgao in filters['orgaos']:
            condition = f'c."Órgão (Código/Nome)" = {orgao}'
            conditions.append(condition)

    query = base_query + ' WHERE ' + ' and '.join(conditions)

    return query



from typing import Any

import tcegoframework.config as config
from pandas import DataFrame, Series
from torch import long, tensor
from torch.utils.data import DataLoader, Dataset


class EmpenhoDataset(Dataset):
    def __init__(self, empenho_historico: Series, target: Series, tokenizer):
        self.empenho_historico = empenho_historico
        self.target = target
        self.tokenizer = tokenizer
        self.max_len = config.BERT_MAX_LEN

    def __len__(self) -> int:
        return len(self.empenho_historico)

    def __getitem__(self, item: int) -> dict[str, Any]:
        empenho_historico = str(self.empenho_historico[item])
        target = self.target[item]
        encoding = self.tokenizer.encode_plus(
            empenho_historico,
            add_special_tokens=True,
            max_length=self.max_len,
            return_token_type_ids=False,
            padding='max_length',
            return_attention_mask=True,
            return_tensors='pt',
            truncation=True
        )
        return {
            'empenho_text': empenho_historico,
            'input_ids': encoding['input_ids'].flatten(),
            'attention_mask': encoding['attention_mask'].flatten(),
            'targets': tensor(target, dtype=long)
        }


def create_data_loader(data: DataFrame, tokenizer) -> DataLoader:

    dataset = EmpenhoDataset(
        empenho_historico=data.empenho_historico.to_numpy(),
        target=data['natureza_despesa_cod'].to_numpy(),
        tokenizer=tokenizer,
    )
    return DataLoader(
        dataset,
        batch_size=config.BERT_BATCH_SIZE,
    )



import time
import warnings
from functools import partial

import tcegoframework.config as config
from pandas.core.frame import DataFrame
from scipy.sparse import csr_matrix
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from tcegoframework.cfgparsing import (get_algorithm, get_epochs,
                                       get_expired_labels_path,
                                       get_label_population_floor,
                                       get_training_dataset_path,
                                       get_training_sampling_number,
                                       get_validated_data_path)
from tcegoframework.data.filter import (change_scope_dict, create_scope_dict,
                                        masked_filter, where_below_threshold,
                                        where_class_92, where_expired_class,
                                        where_zero_value)
from tcegoframework.dremio import (construct_query, execute_query,
                                   get_train_data)
from tcegoframework.io import (dump_model, load_csv_data, load_encoder,
                               load_excel_data, load_torch_model,
                               save_bert_history_plot, save_scope_dict)
from tcegoframework.model.bert import (NaturezaClassifier, fit_bert,
                                       get_predictions)
from tcegoframework.model.metrics import (classification_report_csv,
                                          special_report_csv)
from tcegoframework.preprocessing.classification import (
    preprocessing_training_corretude, preprocessing_training_natureza,
    preprocessing_training_natureza_bert)
from tcegoframework.preprocessing.text import regularize_columns_name

warnings.filterwarnings('ignore')


def dataframe_reset_index(dataframe: DataFrame) -> DataFrame:
    return dataframe.reset_index(drop=True)


def data_cleaning(data: DataFrame) -> tuple[DataFrame, DataFrame]:
    data = regularize_columns_name(data)

    print(f'Dimensões iniciais: {data.shape}')
    print(f'Quantidade de classes: {data.natureza_despesa_cod.unique().shape}')
    print()

    scope_dict = create_scope_dict(data)

    zero_value_mask = where_zero_value(data)
    data, dropped_rows = masked_filter(data, zero_value_mask)
    print()
    print(
        f'Dimensões após remoção de empenhos com saldo zerados: {data.shape}')
    print(
        f'Quantidade de classes restantes: {data.natureza_despesa_cod.unique().shape}')
    print(f'Dimensões dos empenhos removidos: {dropped_rows.shape}')
    print(
        f'Quantidade de classes removidas: {dropped_rows.natureza_despesa_cod.unique().shape}')
    print()

    class92_mask = where_class_92(data)
    data, dropped_rows = masked_filter(data, class92_mask)
    scope_dict = change_scope_dict(
        scope_dict, dropped_rows['natureza_despesa_cod'], 'Classe 92')
    print()
    print(f'Dimensões após remoção de empenhos da Classe 92: {data.shape}')
    print(
        f'Quantidade de classes restantes: {data.natureza_despesa_cod.unique().shape}')
    print(f'Dimensões dos empenhos removidos: {dropped_rows.shape}')
    print(
        f'Quantidade de classes removidas: {dropped_rows.natureza_despesa_cod.unique().shape}')
    print()

    expired_classes = load_excel_data(path=get_expired_labels_path())
    expired_classes = regularize_columns_name(expired_classes)
    expired_classes_mask = where_expired_class(data, expired_classes)
    data, dropped_rows = masked_filter(data, expired_classes_mask)
    scope_dict = change_scope_dict(
        scope_dict, dropped_rows['natureza_despesa_cod'], 'Classe com vigência expirada')
    print()
    print(
        f'Dimensões após remoção de empenhos com vigência encerrada: {data.shape}')
    print(
        f'Quantidade de classes restantes: {data.natureza_despesa_cod.unique().shape}')
    print(f'Dimensões dos empenhos removidos: {dropped_rows.shape}')
    print(
        f'Quantidade de classes removidas: {dropped_rows.natureza_despesa_cod.unique().shape}')
    print()

    threshold = get_label_population_floor()
    under_threshold_mask = where_below_threshold(data, threshold)
    data_above, data_below = masked_filter(data, under_threshold_mask)
    data_below, _ = masked_filter(
        data_below, where_below_threshold(data_below, 2))
    print()
    print(
        f'Dimensões após separação dos empenhos de classes pequenas: {data_above.shape}')
    print(
        f'Quantidade de classes restantes: {data_above.natureza_despesa_cod.unique().shape}')
    print(f'Dimensões dos empenhos de classes pequenas: {data_below.shape}')
    print(
        f'Quantidade de classes pequenas: {data_below.natureza_despesa_cod.unique().shape}')
    print()

    save_scope_dict('scope.pkl', scope_dict)

    return data_above, data_below


def get_dataset(filters: dict) -> DataFrame:
    if dataset_path := get_training_dataset_path():
        data = load_csv_data(dataset_path)
    else:
        if filters:
            query = construct_query(filters)
            data = execute_query(query)
        else:
            data = get_train_data()
    return data


def model_name(type: str, algorithm: str):
    return f'{type}_{algorithm}'


def train_rf_natureza(data: DataFrame, section: str):
    X_train, X_test, y_train, y_test = preprocessing_training_natureza(
        data.copy(), 'tfidf', section)

    model = RandomForestClassifier(
        n_estimators=500, n_jobs=-1, random_state=config.RANDOM_SEED)
    model.fit(X_train, y_train)
    y_pred = model.predict(X_test)

    special_report_csv(
        y_true=y_test,
        y_pred=y_pred,
        data=data,
        encoding=False,
        filename=f'rf_natureza_{section}_rep.csv')
    dump_model(model=model, filename=f'rf_natureza_{section}_model.pkl')


def train_svm_natureza(data: DataFrame, section: str):
    X_train, X_test, y_train, y_test = preprocessing_training_natureza(
        data.copy(), 'tfidf', section)

    model = SVC(kernel='linear', random_state=config.RANDOM_SEED)
    grid = {'C': [0.1, 1, 10, 50]}
    gs = GridSearchCV(model, grid, n_jobs=None, cv=3, verbose=10)
    gs.fit(csr_matrix(X_train.values), y_train)

    model = SVC(C=gs.best_params_['C'], kernel='linear',
                random_state=config.RANDOM_SEED, probability=True)
    model.fit(csr_matrix(X_train.values), y_train)
    y_pred = model.predict(csr_matrix(X_test.values))

    special_report_csv(
        y_true=y_test,
        y_pred=y_pred,
        data=data,
        encoding=False,
        filename=f'svm_natureza_{section}_rep.csv')
    dump_model(model=model, filename=f'svm_natureza_{section}_model.pkl')


def get_n_naturezas(data: DataFrame) -> int:
    n_classes = data['natureza_despesa_cod'].value_counts().shape[0]
    return n_classes


def train_bert_natureza(data: DataFrame, section: str) -> None:
    traindl, testdl = preprocessing_training_natureza_bert(data, section)

    n_classes = get_n_naturezas(data)

    bert_model = NaturezaClassifier(n_classes=n_classes)
    history = fit_bert(model=bert_model, epochs=get_epochs(),
                       train_data_loader=traindl, test_data_loader=testdl, section=section)

    save_bert_history_plot(history, section)

    bert_model = NaturezaClassifier(n_classes=n_classes)
    bert_model = bert_model.to(config.BERT_DEVICE)
    state_dict = load_torch_model(filename=f'bert_model_{section}.bin')
    bert_model.load_state_dict(state_dict)

    result = get_predictions(model=bert_model, data_loader=testdl)

    label_encoder = load_encoder(f'le_bert_{section}.pkl')

    y_true = label_encoder.inverse_transform(result['real_values'])
    y_pred = label_encoder.inverse_transform(result['predictions'])

    special_report_csv(
        y_true=y_true,
        y_pred=y_pred,
        data=data,
        encoding=False,
        filename=f'bert_{section}_rep.csv')


def train_bert_rf_natureza(data: DataFrame, section: str):

    train_bert_natureza(data, section)

    X_train, X_test, y_train, y_test = preprocessing_training_natureza(
        data.copy(), 'bert', section)

    model = RandomForestClassifier(
        n_estimators=500, n_jobs=-1, random_state=config.RANDOM_SEED)
    model.fit(X_train, y_train)
    y_pred = model.predict(X_test)

    special_report_csv(
        y_true=y_test,
        y_pred=y_pred,
        data=data,
        encoding=False,
        filename=f'bert_rf_natureza_{section}_rep.csv')
    dump_model(model=model, filename=f'bert_rf_natureza_{section}_model.pkl')


def train_corretude(data: DataFrame):
    X_train, X_test, y_train, y_test = preprocessing_training_corretude(
        data.copy())

    model = RandomForestClassifier(
        n_estimators=400, random_state=15, n_jobs=-1)
    model.fit(X_train, y_train)
    y_pred = model.predict(X_test)

    classification_report_csv(y_test, y_pred, False, 'rf_corretude_rep.csv')

    dump_model(model=model, filename='rf_corretude_model.pkl')


def train_flow(filters: dict):
    print('Carregando base de dados...')
    data = get_dataset(filters)
    print('Base de dados carregada.')

    if samples := get_training_sampling_number():
        data = data.sample(samples, random_state=config.RANDOM_SEED)
        data = dataframe_reset_index(data.copy())

    print(f'Iniciando limpeza...')
    data_above, data_below = data_cleaning(data.copy())
    print('Limpeza finalizada')

    # BORDA DE TRANSIÇÃO ENTRE LIMPEZA E INÍCIO DO PREPROCESSAMENTO

    if get_algorithm() == 'svm':
        train_natureza = partial(train_svm_natureza)
    elif get_algorithm() == 'rf':
        train_natureza = partial(train_rf_natureza)
    elif get_algorithm() == 'bert_rf':
        train_natureza = partial(train_bert_rf_natureza)

    print('Iniciando treinamento do classificador de natureza de despesa')
    print('Treinamento do modelo de classes maiores')
    time_ref = time.time()
    train_natureza(data_above.copy(), 'above')
    print(
        f'Treinamento finalizado. Duração total: {(time.time() - time_ref)/60}')
    print('Treinamento do modelo de classes menores')
    time_ref = time.time()
    train_natureza(data_below.copy(), 'below')
    print(
        f'Treinamento finalizado. Duração total: {(time.time() - time_ref)/60}')

    # Modelo 2

    data = load_excel_data(get_validated_data_path())
    data = data.reset_index(drop=True)
    data = regularize_columns_name(data)
    print('Iniciando treinamento do classificador de corretude')
    time_ref = time.time()
    train_corretude(data.copy())
    print(
        f'Treinamento finalizado. Duração total: {(time.time() - time_ref)/60}')
    print('Script de treinamento finalizado')



import argparse
import sys

from dateutil.parser import parse

from tcegoframework import config
from tcegoframework.flows.evaluation import evaluation_flow


def parseb(date: str):
    return parse(date, dayfirst=True)


# Define parser
parser = argparse.ArgumentParser(prog='tcegoframework')

# Define args
parser.add_argument(
    'task',
    action='store',
    choices=['training', 'inference', 'evaluation'],
    type=str,
    help='''
        Tarefa: obrigatório. Seleciona a tarefa a ser executada.
        \'training\' Treina o modelo com todos os dados disponíveis até o dia anterior.
        \'inference\' Realiza inferência nos empenhos do dia anterior por padrão (para mais opções de filtro veja outros parâmetros).
        \'evaluation\' Avalia os algoritmos 
        '''
)

group_date = parser.add_mutually_exclusive_group(required=False)

group_date.add_argument(
    '--daterange', '-dr',
    action='store',
    nargs=2,
    type=parseb,
    help='''
        Filtro por intervalo de datas: opcional. 
        Ao usar essa opção é necessário passar duas datas e todos os empenhos da mais antiga até
        (incluindo) a mais recente serão selecionados. O formato de data preferível é DD/MM/AAAA.
        ''',
)

group_date.add_argument(
    '--dates', '-d',
    action='store',
    nargs='*',
    type=parseb,
    help='''
        Filtro por seleção de datas: opcional.
        Podem ser selecionadas datas individuais para fazer a extração dos empenhos. 
        O formato de datas preferível é DD/MM/AAAA.
        ''',
)

parser.add_argument(
    '--orgaos', '-o',
    action='store',
    type=str,
    nargs='*',
    help='''
        Filtrar por orgão(s): opcional. Só é funcional quando a tarefa selecionada é \'infer\'.
        Podem ser selecionados 1 ou mais órgãos por esse filtro.
        ''',
)


def main():
    args = parser.parse_args()
    filters = {}
    if args.task == 'training':
        if args.daterange:
            filters['daterange'] = args.daterange
        from tcegoframework.flows.training import train_flow
        train_flow(filters)
        sys.exit(0)

    elif args.task == 'inference':
        filters = {}
        if args.dates:
            filters['dates'] = args.dates
        elif args.daterange:
            filters['daterange'] = args.daterange
        if args.orgaos:
            filters['orgaos'] = args.orgaos
        from tcegoframework.flows.inference import inference_flow
        inference_flow(filters)
        sys.exit(0)
    elif args.task == 'evaluation':
        evaluation_flow()
        sys.exit(0)



import torch

from tcegoframework.cfgparsing import get_random_seed

# Directiores
ROOT_DIR = 'TCEGO_IA_DATA/'
ENC_PATH = ROOT_DIR + 'bin/encoders/'
MODEL_PATH = ROOT_DIR + 'bin/models/'
STDOUT_REDIR_PATH = ROOT_DIR
META_PATH = ROOT_DIR + 'meta/'

# General
RANDOM_SEED = get_random_seed()

# BERT settings
PRE_TRAINED_MODEL_NAME = 'neuralmind/bert-base-portuguese-cased'
BERT_MAX_LEN = 156
BERT_BATCH_SIZE = 16
BERT_LEARNING_RATE = 2e-5
BERT_DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


# Columns selection
CLF_CAT = [
    'orgao',
    'orgao_sucessor_atual',
    'tipo_administracao_nome',
    'tipo_poder_nome',
    'funcao',
    'subfuncao',
    'programa',
    'acao',
    'modalidade_licitacao_nome',
    'fonte_recurso_nome',
    'beneficiario_cnpj',
    'beneficiario_cpf',
    'beneficiario_cpf/cnpj',
    'periodo',
    'empenho_numero_do_processo',
]
CLF2_CAT = [
    'orgao',
    'orgao_sucessor_atual',
    'tipo_administracao_nome',
    'tipo_poder_nome',
    'funcao',
    'subfuncao',
    'programa',
    'acao',
    'grupo_despesa',
    'natureza_despesa_cod',
    'modalidade_licitacao_nome',
    'fonte_recurso_nome',
    'beneficiario_cnpj',
    'beneficiario_cpf',
    'beneficiario_cpf/cnpj',
    'periodo',
]
CLF_TEXT = [
    'empenho_historico',
]
CLF2_TEXT = [
    'empenho_historico',
]
CLF_NUM = [
    'valor_empenhado',
    'valor_anulacao_empenho',
    'valor_cancelamento_empenho',
    'valor_saldo_do_empenho',
    'valor_liquidacao_empenho',
    'valor_anulacao_liquidacao_empenho',
    'valor_saldo_liquidado',
    'valor_ordem_de_pagamento',
    'valor_guia_recolhimento',
    'valor_anulacao_ordem_de_pagamento',
    'valor_estorno_anulacao_o_pagamento',
    'valor_estorno_guia_recolhimento',
    'valor_saldo_pago',
    'valor_saldo_a_pagar',
    'valor_a_liquidar',
    'valor_a_pagar_liquidado'
]
CLF2_NUM = [
    'valor_empenhado',
    'valor_anulacao_empenho',
    'valor_cancelamento_empenho',
    'valor_saldo_do_empenho',
    'valor_liquidacao_empenho',
    'valor_anulacao_liquidacao_empenho',
    'valor_saldo_liquidado',
    'valor_ordem_de_pagamento',
    'valor_guia_recolhimento',
    'valor_anulacao_ordem_de_pagamento',
    'valor_estorno_anulacao_o_pagamento',
    'valor_estorno_guia_recolhimento',
    'valor_saldo_pago',
    'valor_saldo_a_pagar',
    'valor_a_liquidar',
    'valor_a_pagar_liquidado'
]


# Inference

# Functions to change constants in runtime

def CHANGE_ROOT_DIR(root_dir: str) -> None:
    global ROOT_DIR, ENC_PATH, MODEL_PATH, STDOUT_REDIR_PATH, META_PATH
    ROOT_DIR = root_dir + '/'
    ENC_PATH = ROOT_DIR + 'bin/encoders/'
    MODEL_PATH = ROOT_DIR + 'bin/models/'
    STDOUT_REDIR_PATH = ROOT_DIR
    META_PATH = ROOT_DIR + 'meta/'



from functools import partial
from typing import Iterable

import tcegoframework.config as config
from pandas import DataFrame
from pandas.core import series


def scope_filter(data: DataFrame, scope_dict: dict) -> tuple[DataFrame, DataFrame]:
    mask = data.apply(partial(isinscope, scope_dict), axis=1)
    oos = data.loc[~mask, :].reset_index(drop=True)
    data = data.loc[mask, :].reset_index(drop=True)
    return data, oos


def isinscope(scope_dict: dict, empenho: series.Series) -> bool:
    if empenho['natureza_despesa_cod'] not in scope_dict:
        return False
    elif empenho['natureza_despesa_cod'] in scope_dict:
        key = empenho['natureza_despesa_cod']
        if scope_dict[key] != 'Em escopo':
            return False
        elif empenho['valor_saldo_do_empenho'] == 0:
            return False
        else:
            return True


def create_scope_dict(data: DataFrame) -> dict[str, str]:
    scope_dict = {}
    for target in data['natureza_despesa_cod'].unique():
        scope_dict[target] = 'Em escopo'
    return scope_dict


def change_scope_dict(scope_dict: dict[str, str], target_keys: Iterable[str], scope_status: str) -> dict[str, str]:
    for target_key in target_keys:
        scope_dict[target_key] = scope_status
    return scope_dict


def masked_filter(data: DataFrame, mask: list[bool]) -> tuple[DataFrame, DataFrame]:
    return data.loc[mask, :], data.loc[~mask, :]


def where_class_92(data: DataFrame) -> list[bool]:
    regex = r'\d[.]\d[.]\d\d[.]92[.]\d\d'
    mask = ~data['natureza_despesa_cod'].str.fullmatch(regex)
    return mask


def where_zero_value(data: DataFrame) -> Iterable[bool]:
    return data['valor_saldo_do_empenho'] > 0


def where_expired_class(data: DataFrame, expired_classes: DataFrame) -> Iterable[bool]:
    expired_classes = expired_classes['nat_despesa'].to_list()
    mask = ~data['natureza_despesa_cod'].isin(expired_classes)
    return mask


def where_below_threshold(data: DataFrame, threshold: int) -> Iterable[bool]:
    column = 'natureza_despesa_cod'
    counters = data[column].value_counts().to_frame()
    above_threshold = counters[counters[column] >= threshold].index.to_list()
    mask = data[column].isin(above_threshold)
    return mask



import re
import string

from cleantext import clean
from nltk import download
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from pandas import DataFrame
from unidecode import unidecode


def _clean_partial(text: str) -> str:
    text = clean(text,
                 fix_unicode=True,
                 to_ascii=True,
                 lower=True,
                 normalize_whitespace=True,
                 no_line_breaks=True,
                 strip_lines=True,
                 keep_two_line_breaks=False,
                 no_urls=True,
                 no_emails=True,
                 no_phone_numbers=True,
                 no_numbers=True,
                 no_digits=True,
                 no_currency_symbols=True,
                 no_punct=False,
                 no_emoji=True,
                 replace_with_url="",
                 replace_with_email="",
                 replace_with_phone_number="",
                 replace_with_number="",
                 replace_with_digit="",
                 replace_with_currency_symbol="",
                 replace_with_punct="",
                 lang="pt",)
    return text


def regularize_columns_name(data: DataFrame) -> DataFrame:

    columns = []
    for column_name in data.columns:
        column_name = column_name.lower().replace(' ', '_')
        column_name = column_name.replace('.', '')
        column_name = unidecode(column_name)
        column_name = re.sub('(\(eof\))|\(|\)', '', column_name)
        column_name = re.sub(
            '(_codigo/nome)|(_cod/nome)|(_dia/mes/ano)|', '', column_name)
        columns.append(column_name)
    data.columns = columns
    return data


def stemming(input_text):
    porter = PorterStemmer()
    words = input_text.split()
    stemmed_words = [porter.stem(word) for word in words]
    return " ".join(stemmed_words)


def remove_stopwords(input_text, stopwords_list):
    words = input_text.split()
    clean_words = [word for word in words if (
        word not in stopwords_list) and len(word) > 1]
    return " ".join(clean_words)


def remove_punctuation(input_text):
    # Make translation table
    punct = string.punctuation
    # Every punctuation symbol will be replaced by a space
    trantab = str.maketrans(punct, len(punct)*' ')
    return input_text.translate(trantab)


def clean_nlp(text: str) -> str:
    text = remove_punctuation(text)
    return _clean_partial(text)


def clean_tfidf(text: str) -> str:

    text = remove_punctuation(text)
    text = _clean_partial(text,)

    try:
        stopwords_list = stopwords.words('portuguese')
    except:
        download("stopwords")
        download('wordnet')
        stopwords_list = stopwords.words('portuguese')

    stopwords_list.append("pdf")
    stopwords_list.append("total")
    stopwords_list.append("mes")
    stopwords_list.append("goias")
    stopwords_list.append("go")
    text = remove_stopwords(text, stopwords_list)
    text = stemming(text)
    return text



from configparser import ConfigParser
from typing import Union

PARSER = ConfigParser()
PARSER.read('config.ini')

GENERAL = 'options.general'
DREMIO = 'options.dremio'
TRAINING = 'options.training'
INFERENCE = 'options.inference'
HDFS = 'options.hdfs'


# [options.general]

def get_random_seed() -> int:
    return PARSER.getint(GENERAL, 'random_seed', fallback=15)


# [options.dremio]


def get_dremio_connection() -> str:
    return PARSER.get(DREMIO, 'connection')


def get_dremio_user() -> str:
    return PARSER.get(DREMIO, 'user')


def get_dremio_password() -> str:
    return PARSER.get(DREMIO, 'password')


# [options.training]


def get_training_dataset_path() -> Union[str, None]:
    return PARSER.get(TRAINING, 'dataset_path', fallback=None)


def get_training_sampling_number() -> Union[int, None]:
    return PARSER.getint(TRAINING, 'sample_dataset', fallback=None)


def get_expired_labels_path() -> str:
    return PARSER.get(TRAINING, 'expired_class_path')


def get_validated_data_path() -> str:
    return PARSER.get(TRAINING, 'validated_data_path')


def get_label_population_floor() -> int:
    return PARSER.getint(TRAINING, 'min_documents_class', fallback=200)


def get_algorithm() -> str:
    return PARSER.get(TRAINING, 'algorithm', fallback='bert_rf')


def get_epochs() -> int:
    return PARSER.getint(TRAINING, 'epochs', fallback=5)

# [options.inference]


def get_inference_dataset_path() -> Union[str, None]:
    return PARSER.get(INFERENCE, 'dataset_path', fallback=None)


# [options.hdfs]

def use_hdfs() -> bool:
    return PARSER.getboolean(HDFS, 'use', fallback=True)


def get_hdfs_domain() -> str:
    return PARSER.get(HDFS, 'domain')


def get_hdfs_url() -> str:
    return PARSER.get(HDFS, 'url')


def get_hdfs_port() -> int:
    return PARSER.getint(HDFS, 'port')


def get_hdfs_user() -> str:
    return PARSER.get(HDFS, 'user')


def get_hdfs_password() -> str:
    return PARSER.get(HDFS, 'password')


def get_hdfs_dir_path() -> str:
    return PARSER.get(HDFS, 'path')



import tcegoframework.config as config
from pandas import DataFrame


def initialize_inference_dict(data: DataFrame) -> DataFrame:
    config.INFERENCE_DICT = {}
    for i in range(0, data.shape[0]):
        empenho = data.iloc[i, :]
        config.INFERENCE_DICT[empenho['empenho_sequencial_empenho']] = {
            'Identificador': empenho['empenho_sequencial_empenho'],
            'Natureza Real': empenho['natureza_despesa_cod'],
            'Natureza Predita': None,
            'Corretude': None,
            'Resultado': None,
        }


def resolve_output(data, y1, y2):
    for i in range(data.shape[0]):
        key = data.iloc[i, :]['empenho_sequencial_empenho']
        config.INFERENCE_DICT[key]['Natureza Predita'] = y1[i]
        config.INFERENCE_DICT[key]['Corretude'] = y2[i]
        config.INFERENCE_DICT[key]['Resultado'] = resolve_result(
            config.INFERENCE_DICT[key]['Natureza Real'],
            config.INFERENCE_DICT[key]['Natureza Predita'],
            config.INFERENCE_DICT[key]['Corretude']
        )


def resolve_result(y_true, y_pred, y_pred_correctness) -> str:

    if y_true == y_pred:
        if y_pred_correctness == 'INCORRETO':
            return 'INCV_M2'
        elif y_pred_correctness == 'OK':
            return 'C_M1-M2'
        elif y_pred_correctness == 'INCONCLUSIVO':
            return 'AD_M2'
    else:
        if y_pred_correctness == 'INCORRETO':
            return 'INCT_M1-M2'
        elif y_pred_correctness == 'OK':
            return 'INCV_M1'
        elif y_pred_correctness == 'INCONCLUSIVO':
            return 'INCV_M1-AD_M2'



from pandas import DataFrame
from sklearn.metrics import classification_report
from tcegoframework.io import load_encoder, save_csv_data


def classification_report_csv(y_true, y_pred, encoding: bool, filename: str) -> None:
    if encoding:
        encoder = load_encoder('targetNLP.pkl')
        report = classification_report(
            y_true, y_pred, target_names=encoder.classes_, output_dict=True)
    else:
        report = classification_report(y_true, y_pred, output_dict=True)
    report = DataFrame(report).transpose()
    save_csv_data(report, filename)


def special_report_csv(y_true, y_pred, data: DataFrame, encoding: bool, filename: str) -> None:
    if encoding:
        encoder = load_encoder('targetNLP.pkl')
        report = classification_report(
            y_true, y_pred, target_names=encoder.classes_, output_dict=True)
    else:
        report = classification_report(y_true, y_pred, output_dict=True)
    value_counts = data['natureza_despesa_cod'].value_counts().to_dict()
    for key in report.keys():
        if key not in ['accuracy', 'macro avg', 'weighted avg']:
            report[key]['n docs'] = value_counts[key]
            report[key]['percent of total'] = value_counts[key] / data.shape[0]
        elif key == 'accuracy':
            continue
        else:
            report[key]['n docs'] = data.shape[0]
            report[key]['percent of total'] = 1
    report = DataFrame(report).transpose()
    save_csv_data(report, filename)



from datetime import datetime

from dateutil.parser import parse
from numpy import array, zeros
from pandas import DataFrame


def get_n_classes(data: DataFrame, column: str) -> int:
    n_classes = data[column].value_counts().shape[0]
    return n_classes


def code_reaper(input_text):
    return input_text[7:]


def date_to_month(input_date):
    if isinstance(input_date, int):
        date = datetime.fromordinal(
            datetime(1900, 1, 1).toordinal() +
            input_date -
            2)
    else:
        date = parse(input_date)
    return date.month


def data_preparation(data: DataFrame, categorical_columns, numerical_columns):

    data = data.reset_index(drop=True)
    # Criação de meta-atributos e tratamento comum para todos os dados

    # Criação do meta-atributo mês
    if 'periodo' in categorical_columns:
        data['periodo'].update(data['periodo'].map(date_to_month))

    # Criação do meta-atributo "Pessoa Jurídica?"
    pessoa_juridica = array(
        [1 if cpf == '-' else 0
         for cpf in data['beneficiario_cpf'].values])
    data['pessoa_juridica'] = pessoa_juridica.astype('int8')

    data.drop(['beneficiario_cpf', 'beneficiario_cnpj',
               'beneficiario_cpf/cnpj'], axis='columns', inplace=True)
    categorical_columns.remove('beneficiario_cpf')
    categorical_columns.remove('beneficiario_cnpj')
    categorical_columns.remove('beneficiario_cpf/cnpj')
    categorical_columns.append('pessoa_juridica')
    del pessoa_juridica

    # Codigo que gera o meta atributo "orgao_sucedido" onde 1 representa que o orgao tem um novo orgao sucessor e 0 caso contrario
    orgao_sucedido = zeros(data.shape[0])

    for i in range(data.shape[0]):
        if(data['orgao'].iloc[i] != data['orgao_sucessor_atual'].iloc[i]):
            orgao_sucedido[i] = 1

    data['orgao_sucedido'] = orgao_sucedido.astype('int8')
    data.drop(['orgao'], axis='columns', inplace=True)
    categorical_columns.remove('orgao')
    categorical_columns.append('orgao_sucedido')
    del orgao_sucedido

    # Codigo que retira o codigo de programa (retirando 10 valores)
    data['programa'] = data['programa'].map(code_reaper)
    # Codigo que retira o codigo de acao (retirando 77 valores)
    data['acao'] = data['acao'].map(code_reaper)

    # Codigo que concatena acao e programa
    acao_programa = zeros(data.shape[0], dtype='object')
    for i in range(data.shape[0]):
        acao_programa[i] = (data['acao'].iloc[i] + ' & ' +
                            data['programa'].iloc[i])
    data['acao_programa'] = acao_programa
    data.drop(['acao', 'programa'], axis='columns', inplace=True)
    categorical_columns.remove('acao')
    categorical_columns.remove('programa')
    categorical_columns.append('acao_programa')
    del acao_programa

    return data, categorical_columns, numerical_columns



from copy import deepcopy

from pandas import DataFrame
from pandas.core.series import Series
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from tcegoframework import config
from tcegoframework.data.misc import create_data_loader
from tcegoframework.io import dump_encoder, load_encoder
from tcegoframework.preprocessing.encoders import (encode_inference,
                                                   encode_train_test)
from tcegoframework.preprocessing.text import clean_nlp
from tcegoframework.preprocessing.transform import data_preparation
from torch.utils.data import dataloader
from transformers import BertTokenizer


def preprocessing_training_natureza(data: DataFrame, text_representation: str, section: str) -> tuple[
        DataFrame, DataFrame, Series, Series]:
    target = data.natureza_despesa_cod

    cat_col = config.CLF_CAT.copy()
    text_col = config.CLF_TEXT.copy()
    num_col = config.CLF_NUM.copy()

    data = data.loc[:, (*cat_col,
                        *text_col,
                        *num_col,)]

    data, categorical_columns, numerical_columns = data_preparation(
        data,
        categorical_columns=cat_col,
        numerical_columns=num_col, )

    X_train, X_test, y_train, y_test = train_test_split(
        data, target, test_size=0.3, random_state=config.RANDOM_SEED, stratify=target)

    X_train, X_test = encode_train_test(X_train,
                                        X_test,
                                        numerical_columns,
                                        categorical_columns,
                                        text_col,
                                        y_train=y_train,
                                        y_test=y_test,
                                        prefix=f'nat_{section}',
                                        text_representation=text_representation,
                                        section=section)

    return X_train, X_test, y_train, y_test


def preprocessing_inference_natureza(data: DataFrame, text_representation: str, section: str) -> DataFrame:
    target = data.natureza_despesa_cod

    cat_col = [
        # 'exercicio_do_orcamento_ano',
        # 'empenho_sequencial_empenho',
        'orgao',
        'orgao_sucessor_atual',
        'tipo_administracao_nome',
        'tipo_poder_nome',
        # 'classificacao_orcamentaria_descricao',
        'funcao',
        'subfuncao',
        'programa',
        'acao',
        # 'grupo_despesa',
        # 'elemento_despesa',
        # 'natureza_despesa_cod',
        # 'natureza_despesa_nome',
        # 'formalidade_nome',
        'modalidade_licitacao_nome',
        # 'fonte_recurso_cod',
        'fonte_recurso_nome',
        'beneficiario_cnpj',
        'beneficiario_cpf',
        'beneficiario_cpf/cnpj',
        'periodo',
        'empenho_numero_do_processo',
        # 'empenho_sequencial_empenho.1',
    ]
    text_col = [
        # 'beneficiario_nome',
        'empenho_historico',
    ]
    num_col = [
        'valor_empenhado',
        'valor_anulacao_empenho',
        # 'valor_estorno_anulacao_empenho',
        'valor_cancelamento_empenho',
        # 'valor_anulacao_cancelamento_empenho',
        'valor_saldo_do_empenho',
        'valor_liquidacao_empenho',
        'valor_anulacao_liquidacao_empenho',
        'valor_saldo_liquidado',
        'valor_ordem_de_pagamento',
        'valor_guia_recolhimento',
        'valor_anulacao_ordem_de_pagamento',
        'valor_estorno_anulacao_o_pagamento',
        'valor_estorno_guia_recolhimento',
        'valor_saldo_pago',
        'valor_saldo_a_pagar',
        'valor_a_liquidar',
        'valor_a_pagar_liquidado'
    ]

    data = data.loc[:, (*cat_col,
                        *text_col,
                        *num_col,)]

    data, categorical_columns, numerical_columns = data_preparation(
        data, categorical_columns=cat_col, numerical_columns=config.CLF_NUM, )

    X = encode_inference(
        data,
        target,
        numerical_columns,
        categorical_columns,
        config.CLF_TEXT,
        prefix=f'nat_{section}',
        text_representation=text_representation,
        section=section)

    return X


def preprocessing_training_natureza_bert(data: DataFrame, section: str) -> tuple[dataloader.DataLoader, dataloader.DataLoader]:
    data = data[['empenho_historico', 'natureza_despesa_cod']]
    data['empenho_historico'].update(data['empenho_historico'].map(clean_nlp))
    tokenizer = BertTokenizer.from_pretrained(config.PRE_TRAINED_MODEL_NAME)

    df_train, df_test = train_test_split(
        data,
        test_size=0.3,
        random_state=config.RANDOM_SEED,
        stratify=data['natureza_despesa_cod']
    )

    df_train = df_train.reset_index(drop=True)
    df_test = df_test.reset_index(drop=True)

    encoder = LabelEncoder()
    df_train['natureza_despesa_cod'].update(
        encoder.fit_transform(df_train['natureza_despesa_cod']))
    df_test['natureza_despesa_cod'].update(
        encoder.transform(df_test['natureza_despesa_cod']))
    dump_encoder(encoder, f'le_bert_{section}.pkl')

    traindl = create_data_loader(df_train, tokenizer)
    testdl = create_data_loader(df_test, tokenizer)

    return traindl, testdl


def preprocessing_inference_natureza_bert(data: DataFrame, section: str) -> dataloader.DataLoader:
    data = data[['empenho_historico', 'natureza_despesa_cod']]
    data['empenho_historico'].update(data['empenho_historico'].map(clean_nlp))
    tokenizer = BertTokenizer.from_pretrained(config.PRE_TRAINED_MODEL_NAME)

    encoder = load_encoder(f'le_bert_{section}.pkl')
    data['natureza_despesa_cod'].update(
        encoder.fit_transform(data['natureza_despesa_cod']))

    dataloader = create_data_loader(data, tokenizer)

    return dataloader


def preprocessing_training_corretude(data: DataFrame) -> tuple[DataFrame, DataFrame, Series, Series]:
    target = data.analise

    data = data.loc[:, (*config.CLF2_CAT,
                        *config.CLF2_TEXT,
                        *config.CLF2_NUM,)]

    data, categorical_columns, numerical_columns = data_preparation(
        data,
        categorical_columns=config.CLF2_CAT,
        numerical_columns=config.CLF2_NUM, )

    X_train, X_test, y_train, y_test = train_test_split(
        data, target, test_size=0.2, random_state=config.RANDOM_SEED, stratify=target)

    X_train, X_test = encode_train_test(X_train,
                                        X_test,
                                        numerical_columns,
                                        categorical_columns,
                                        config.CLF2_TEXT,
                                        y_train=y_train,
                                        y_test=y_test,
                                        prefix='cr',
                                        text_representation='tfidf')

    return X_train, X_test, y_train, y_test


def preprocessing_inference_corretude(data: DataFrame) -> DataFrame:
    data = data.loc[:, (*config.CLF2_CAT,
                        *config.CLF2_TEXT,
                        *config.CLF2_NUM,)]

    data, categorical_columns, numerical_columns = data_preparation(
        data,
        categorical_columns=config.CLF2_CAT,
        numerical_columns=config.CLF2_NUM, )

    X = encode_inference(
        data,
        None,
        numerical_columns,
        categorical_columns,
        config.CLF2_TEXT,
        prefix='cr',
        text_representation='tfidf')

    return X



import time
from functools import partial

from numpy import array
from pandas.core.frame import DataFrame
from sklearn.metrics import classification_report
from tcegoframework.cfgparsing import get_algorithm, get_validated_data_path
from tcegoframework.flows.inference import (inference_bert_rf_natureza,
                                            inference_rf_natureza,
                                            inference_svm_natureza)
from tcegoframework.io import load_excel_data
from tcegoframework.preprocessing.text import regularize_columns_name


def compute_agreement(y_true, y_pred) -> str:
    if y_true == y_pred:
        return 'OK'
    else:
        return 'INCONCLUSIVO'


def compute_output(data: DataFrame, inference_dict: dict, y_natureza: array, y_corretude: array) -> dict:
    for i in range(data.shape[0]):
        pass
    return inference_dict


def evaluation_flow():
    # Executing query
    print('Preparando e executando avaliação...')
    data = load_excel_data(get_validated_data_path())
    data = data.reset_index(drop=True)
    data = regularize_columns_name(data)

    if get_algorithm() == 'svm':
        inference_natureza = partial(inference_svm_natureza)
    elif get_algorithm() == 'rf':
        inference_natureza = partial(inference_rf_natureza)
    elif get_algorithm() == 'bert_rf':
        inference_natureza = partial(inference_bert_rf_natureza)

    for label in ['OK', 'INCONCLUSIVO', 'INCORRETO']:
        temp_data = data.loc[data.analise == label]

        print(f'Inferência de Natureza para documentos {label}...')
        time_ref = time.time()
        y_pred_natureza = inference_natureza(temp_data.copy())

        y_true = temp_data.natureza_despesa_cod

        report = classification_report(
            y_true, y_pred_natureza, output_dict=True)
        report = DataFrame(report).transpose()
        report.to_csv(f'{label}_eval_report.csv')
        print(f'Duração total: {(time.time() - time_ref)/60}')

    print('Finalizado.')



import time
from collections import defaultdict
from typing import Any

import numpy as np
import pandas as pd
import tcegoframework.config as config
import torch
from pandas import DataFrame
from sklearn.metrics import f1_score
from tcegoframework.io import load_torch_model, save_csv_data, save_torch_model
from torch import nn
from torch.utils.data import DataLoader
from transformers import (AdamW, BertModel, get_linear_schedule_with_warmup,
                          logging)

logging.set_verbosity_error()


class NaturezaClassifier(nn.Module):
    def __init__(self, n_classes: int, pre_trained_model_name: str = config.PRE_TRAINED_MODEL_NAME):
        super(NaturezaClassifier, self).__init__()
        self.bert = BertModel.from_pretrained(pre_trained_model_name)
        self.drop = nn.Dropout(p=0.3)
        self.out = nn.Linear(self.bert.config.hidden_size, n_classes)

    def forward(self, input_ids, attention_mask):
        bert_output = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask
        )

        output = self.drop(bert_output['pooler_output'])
        return self.out(output), bert_output['pooler_output']


def train_epoch(model: NaturezaClassifier, data_loader, loss_fn, optimizer, scheduler, n_examples):
    model = model.train()
    losses = []
    correct_predictions = 0
    predictions = []
    real_values = []
    for d in data_loader:
        input_ids = d["input_ids"].to(config.BERT_DEVICE)
        attention_mask = d["attention_mask"].to(config.BERT_DEVICE)
        targets = d["targets"].to(config.BERT_DEVICE)
        outputs, _ = model(
            input_ids=input_ids,
            attention_mask=attention_mask
        )
        _, preds = torch.max(outputs, dim=1)
        loss = loss_fn(outputs, targets)
        correct_predictions += torch.sum(preds == targets)
        losses.append(loss.item())
        loss.backward()
        nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
        optimizer.step()
        scheduler.step()
        optimizer.zero_grad()
        predictions.extend(preds)
        real_values.extend(targets)
    predictions = torch.stack(predictions).cpu()
    real_values = torch.stack(real_values).cpu()
    macro = f1_score(real_values, predictions, average='macro')
    micro = f1_score(real_values, predictions, average='micro')
    return correct_predictions.double() / n_examples, np.mean(losses), macro, micro


def eval_model(model, data_loader, loss_fn, n_examples):
    model = model.eval()
    losses = []
    correct_predictions = 0
    predictions = []
    real_values = []
    with torch.no_grad():
        for d in data_loader:
            input_ids = d["input_ids"].to(config.BERT_DEVICE)
            attention_mask = d["attention_mask"].to(config.BERT_DEVICE)
            targets = d["targets"].to(config.BERT_DEVICE)
            outputs, _ = model(
                input_ids=input_ids,
                attention_mask=attention_mask
            )
            _, preds = torch.max(outputs, dim=1)
            loss = loss_fn(outputs, targets)
            correct_predictions += torch.sum(preds == targets)
            losses.append(loss.item())
            predictions.extend(preds)
            real_values.extend(targets)
    predictions = torch.stack(predictions).cpu()
    real_values = torch.stack(real_values).cpu()
    macro = f1_score(real_values, predictions, average='macro')
    micro = f1_score(real_values, predictions, average='micro')
    return correct_predictions.double() / n_examples, np.mean(losses), macro, micro


def fit_bert(model: NaturezaClassifier, epochs: int, train_data_loader: DataLoader, test_data_loader: DataLoader, section: str) -> dict:

    # np.random.seed(RANDOM_SEED)
    # torch.manual_seed(RANDOM_SEED)
    # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model = model.to(config.BERT_DEVICE)
    optimizer = AdamW(model.parameters(), lr=2e-5, correct_bias=False)
    total_steps = len(train_data_loader) * epochs
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=0,
        num_training_steps=total_steps
    )
    loss_fn = nn.CrossEntropyLoss().to(config.BERT_DEVICE)

    history = defaultdict(list)
    best_accuracy = 0
    for epoch in range(epochs):
        starting = time.time()
        print(f'Epoch {epoch + 1}')
        print('-' * 10)
        train_acc, train_loss, train_macro, train_micro = train_epoch(
            model,
            train_data_loader,
            loss_fn,
            optimizer,
            scheduler,
            len(train_data_loader.dataset)
        )
        print(
            f'Train loss {train_loss} macro {train_macro} micro {train_micro}')
        val_acc, val_loss, val_macro, val_micro = eval_model(
            model,
            test_data_loader,
            loss_fn,
            len(test_data_loader.dataset)
        )
        print(
            f'Val   loss {val_loss} macro {val_macro} micro {val_micro}')
        print('',)
        # history['train_acc'].append(train_acc)
        history['train_loss'].append(train_loss)
        history['train_macro'].append(train_macro)
        history['train_micro'].append(train_micro)
        # history['val_acc'].append(val_acc)
        history['val_loss'].append(val_loss)
        history['val_macro'].append(val_macro)
        history['val_micro'].append(val_micro)

        if val_acc > best_accuracy:
            save_torch_model(model.state_dict(), f'bert_model_{section}.bin')
            best_accuracy = val_acc

        print(
            f'Epoch time: {(time.time()-starting)/60}')

    history = DataFrame(history)
    save_csv_data(history, f'bert_history_{section}.csv')

    return history


def get_predictions(model: NaturezaClassifier, data_loader: DataLoader,) -> dict:
    model = model.to(config.BERT_DEVICE)
    model = model.eval()
    review_texts = []
    predictions = []
    prediction_probs = []
    real_values = []
    with torch.no_grad():
        for d in data_loader:
            texts = d["empenho_text"]
            input_ids = d["input_ids"].to(config.BERT_DEVICE)
            attention_mask = d["attention_mask"].to(config.BERT_DEVICE)
            targets = d["targets"].to(config.BERT_DEVICE)
            outputs, _ = model(
                input_ids=input_ids,
                attention_mask=attention_mask
            )
            _, preds = torch.max(outputs, dim=1)
            review_texts.extend(texts)
            predictions.extend(preds)
            prediction_probs.extend(outputs)
            real_values.extend(targets)
    predictions = torch.stack(predictions).cpu()
    prediction_probs = torch.stack(prediction_probs).cpu()
    real_values = torch.stack(real_values).cpu()
    return {'review_texts': review_texts, 'predictions': predictions, 'prediction_probs': prediction_probs, 'real_values': real_values}


def generate_bert_representation(model, data_loader) -> DataFrame:
    model = model.to(config.BERT_DEVICE)
    model = model.eval()
    with torch.no_grad():
        outs = []
        for d in data_loader:
            _ = d['empenho_text']
            input_ids = d['input_ids'].to(config.BERT_DEVICE)
            attention_mask = d['attention_mask'].to(config.BERT_DEVICE)
            _, pooler = model(input_ids, attention_mask)
            outs.extend(pooler)

    representation = pd.DataFrame(
        np.array([tensor.cpu().detach().numpy() for tensor in outs]),
        columns=[f'BERT_{n}' for n in range(0, np.array([tensor.cpu().detach().numpy() for tensor in outs]).shape[1])])

    return representation


def get_saved_model(n_classes: int, section: str):
    model = NaturezaClassifier(
        n_classes, config.PRE_TRAINED_MODEL_NAME)
    state_dict = load_torch_model(f'bert_model_{section}.bin')
    model.load_state_dict(state_dict)
    return model



import json
from os import makedirs, rename
from os.path import dirname
from typing import Any, Union

import dill
import joblib
import matplotlib.pyplot as plt
import seaborn as sns
from pandas import DataFrame, read_csv, read_excel
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from torch import device, load, save

import tcegoframework.config as config
from tcegoframework.cfgparsing import (get_hdfs_dir_path, get_hdfs_domain,
                                       get_hdfs_password, get_hdfs_port,
                                       get_hdfs_url, get_hdfs_user, use_hdfs)
from tcegoframework.tcehdfs import TCEHDFS


def load_csv_data(path: str) -> DataFrame:
    data = read_csv(path, encoding='utf-8', low_memory=False)
    return data


def save_csv_data(data: DataFrame, filename: str):
    filename = config.META_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    data.to_csv(filename)


def load_excel_data(path: str) -> DataFrame:
    data = read_excel(path)
    return data


def dump_encoder(encoder: Any, filename: str) -> None:
    filename = config.ENC_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    with open(filename, 'wb') as file:
        dill.dump(encoder, file)
        file.close()


def load_encoder(filename: str) -> Any:
    filename = config.ENC_PATH + filename
    with open(filename, 'rb') as file:
        encoder = dill.load(file)
        file.close()
    return encoder


def save_torch_model(modelstate: Any, filename: str) -> None:
    filename = config.MODEL_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    save(modelstate, filename)


def load_torch_model(filename: str, ) -> Any:
    filename = config.MODEL_PATH + filename
    state = load(filename, map_location=config.BERT_DEVICE)
    return state


def dump_model(model: Any, filename: str) -> None:
    filename = config.MODEL_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    joblib.dump(model, filename, compress=3)
    # with open(filename, 'wb') as file:
    #     dill.dump(model, file)
    #     file.close()


def load_model(filename: str) -> Union[SVC, RandomForestClassifier]:
    filename = config.MODEL_PATH + filename
    model = joblib.load(filename)
    # with open(filename, 'rb') as file:
    #     model = dill.load(file)
    #     file.close()
    return model


def printfile(content: str, filename: str = 'out.txt') -> None:
    filename = config.STDOUT_REDIR_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    with open(filename, 'a') as file:
        print(content, file=file)
        file.close()


def change_root_dir_name(directory_name):
    rename(config.ROOT_DIR, directory_name)


def save_scope_dict(filename: str, scope_dict: dict[str, str]) -> None:
    filename = config.MODEL_PATH + filename
    makedirs(dirname(filename), exist_ok=True)
    with open(filename, 'wb') as file:
        dill.dump(scope_dict, file)
        file.close()


def load_scope_dict(filename):
    filename = config.MODEL_PATH + filename
    with open(filename, 'rb') as file:
        class_dict = dill.load(file)
        file.close()
    return class_dict


def save_inference_results(filename: str, inference_dict: dict) -> DataFrame:
    result = DataFrame(inference_dict).transpose()
    result.to_csv(filename)

    if use_hdfs():
        try:
            hdfs = TCEHDFS(
                get_hdfs_domain(),
                get_hdfs_url(),
                get_hdfs_port(),
                get_hdfs_user(),
                get_hdfs_password()
            )

            hdfs.writeHDFS(
                get_hdfs_dir_path(),
                filename,
                result
            )
        except Exception as e:
            print('Ocorreu um erro ao salvar os resultados no HDFS. Erro: ', e)

    return result


def save_inference_plot(filename: str, data: DataFrame) -> None:
    sns.set()
    sns.set_context('talk')
    plt.figure(figsize=(15, 10))
    sns.countplot(y='Resultado', data=data)
    plt.title('Resumo dos resultados')
    plt.savefig(filename)


def save_bert_history_plot(history: dict, section: str) -> None:

    sns.set()
    sns.set_context('talk')
    plt.figure(figsize=(15, 10))

    plt.cla()
    plt.plot(history['train_loss'], label='train loss')
    plt.plot(history['val_loss'], label='validation loss')
    plt.title('Training history')
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend()
    plt.ylim([0, 1])
    plt.savefig(f'{config.META_PATH}/{section}_loss.png')

    plt.cla()
    plt.plot(history['train_macro'], label='train macro')
    plt.plot(history['val_macro'], label='validation macro')
    plt.title('Training history')
    plt.ylabel('Macro')
    plt.xlabel('Epoch')
    plt.legend()
    plt.ylim([0, 1])
    plt.savefig(f'{config.META_PATH}/{section}_macro.png')

    plt.cla()
    plt.plot(history['train_micro'], label='train micro')
    plt.plot(history['val_micro'], label='validation micro')
    plt.title('Training history')
    plt.ylabel('Micro')
    plt.xlabel('Epoch')
    plt.legend()
    plt.ylim([0, 1])
    plt.savefig(f'{config.META_PATH}/{section}_micro.png')


def save_json(data, filename):
    with open(filename, 'w') as file:
        json.dump(data, file)
        file.close()



from numpy import zeros
from pandas import DataFrame
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
from tcegoframework import config
from tcegoframework.data.misc import create_data_loader
from tcegoframework.io import dump_encoder, load_encoder
from tcegoframework.model.bert import (generate_bert_representation,
                                       get_saved_model)
from tcegoframework.preprocessing.text import clean_nlp, clean_tfidf
from transformers import BertTokenizer


def generate_fit_scaler(X_train, columns, prefix=''):
    for col_name in columns:
        scaler = MinMaxScaler()
        X_train[col_name].update(scaler.fit_transform(
            X_train[col_name].values.reshape(-1, 1)).flatten())
        dump_encoder(scaler, f'{prefix}_sc_{col_name}.pkl')
        del scaler
    return X_train


def generate_scaler(X_test, columns, prefix=''):
    for col_name in columns:
        scaler = load_encoder(f'{prefix}_sc_{col_name}.pkl')
        X_test[col_name].update(scaler.transform(
            X_test[col_name].values.reshape(-1, 1)).flatten())
    return X_test


def generate_fit_ohe(X_train, columns, prefix=''):
    for col_name in columns:
        ohe = OneHotEncoder(handle_unknown='ignore')
        enc = ohe.fit_transform(X_train[col_name].values.reshape(-1, 1))
        enc = DataFrame.sparse.from_spmatrix(
            enc, columns=ohe.get_feature_names(input_features=(col_name,)))
        X_train.drop([col_name], inplace=True, axis='columns')
        X_train = X_train.join(enc)
        dump_encoder(ohe, f'{prefix}_ohe_{col_name}.pkl')
        del ohe
    return X_train


def generate_ohe(X_test, columns, prefix=''):
    for col_name in columns:
        ohe = load_encoder(f'{prefix}_ohe_{col_name}.pkl')
        enc = ohe.transform(X_test[col_name].values.reshape(-1, 1))
        enc = DataFrame.sparse.from_spmatrix(
            enc, columns=ohe.get_feature_names(input_features=(col_name,)))
        X_test.drop([col_name], inplace=True, axis='columns')
        X_test = X_test.join(enc)
    return X_test


def generate_fit_tfidf(X_train, columns, prefix=''):
    for col_name in columns:
        X_train[col_name].update(X_train[col_name].map(clean_tfidf))
        tfv = TfidfVectorizer()
        tfidf = tfv.fit_transform(X_train[col_name])
        columns_names = ['Tfidf_' +
                         word for word in tfv.get_feature_names()]
        tfidf = DataFrame.sparse.from_spmatrix(tfidf, columns=columns_names)
        X_train.drop([col_name], inplace=True, axis='columns')
        X_train = X_train.join(tfidf)
        dump_encoder(tfv, f'{prefix}_tfv_{col_name}.pkl')
        del tfv
    return X_train


def generate_tfidf(X_test, columns, prefix=''):
    for col_name in columns:
        X_test[col_name].update(X_test[col_name].map(clean_tfidf))
        tfv = load_encoder(f'{prefix}_tfv_{col_name}.pkl')
        tfidf = tfv.transform(X_test[col_name])
        columns_names = ['Tfidf_' +
                         word for word in tfv.get_feature_names()]
        tfidf = DataFrame.sparse.from_spmatrix(tfidf, columns=columns_names)
        X_test.drop([col_name], inplace=True, axis='columns')
        X_test = X_test.join(tfidf)
    return X_test


def generate_fit_process_count(data):
    process_count = data['empenho_numero_do_processo'].value_counts()
    process_count = process_count.to_dict()

    empenhos_processo = zeros(data.shape[0])

    for i in range(data.shape[0]):
        empenhos_processo[i] = process_count[data['empenho_numero_do_processo'].iloc[i]]
    data['empenhos_por_processo'] = empenhos_processo
    data.drop('empenho_numero_do_processo', axis='columns', inplace=True)
    dump_encoder(process_count, f'enc_empenhos_por_processo.pkl')

    return data


def generate_process_count(data):
    process_count = load_encoder(f'enc_empenhos_por_processo.pkl')

    empenhos_processo = zeros(data.shape[0])

    for i in range(data.shape[0]):
        key = data['empenho_numero_do_processo'].iloc[i]
        if key in process_count:
            empenhos_processo[i] = process_count[key]
        else:
            empenhos_processo[i] = 0
    data['empenhos_por_processo'] = empenhos_processo

    data.drop('empenho_numero_do_processo', axis='columns', inplace=True)

    return data


def generate_bert(X, y, columns, model, encoder) -> DataFrame:
    tokenizer = BertTokenizer.from_pretrained(config.PRE_TRAINED_MODEL_NAME)
    for col_name in columns:
        X[col_name].update(X[col_name].map(clean_nlp))
        data = DataFrame()
        data[col_name] = X[col_name].reset_index(drop=True)
        data['natureza_despesa_cod'] = y.reset_index(drop=True)
        data['natureza_despesa_cod'].update(
            zeros(shape=data['natureza_despesa_cod'].shape))
        data_loader = create_data_loader(
            data, tokenizer)
        representation = generate_bert_representation(model, data_loader)
        X.drop([col_name], inplace=True, axis='columns')
        X = X.join(representation)
    return X


def encode_train_test(X_train: DataFrame, X_test: DataFrame, numerical_columns: list, categorical_columns: list,
                      text_columns: list, y_train, y_test, prefix: str, text_representation: str, section: str = None) -> DataFrame:
    X_train = X_train.reset_index(drop=True)
    X_test = X_test.reset_index(drop=True)

    if 'empenho_numero_do_processo' in categorical_columns:
        X_train = generate_fit_process_count(X_train)
        X_test = generate_process_count(X_test)
        categorical_columns.remove('empenho_numero_do_processo')
        numerical_columns.append('empenhos_por_processo')

    X_train = generate_fit_scaler(X_train, numerical_columns, prefix)
    X_test = generate_scaler(X_test, numerical_columns, prefix)

    X_train = generate_fit_ohe(X_train, categorical_columns, prefix)
    X_test = generate_ohe(X_test, categorical_columns, prefix)

    if text_representation == 'tfidf':
        X_train = generate_fit_tfidf(X_train, text_columns, prefix)
        X_test = generate_tfidf(X_test, text_columns, prefix)
    elif text_representation == 'bert':
        label_encoder = load_encoder(f'le_bert_{section}.pkl')
        model = get_saved_model(len(label_encoder.classes_), section)
        X_train = generate_bert(
            X_train, y_train, text_columns, model, label_encoder)
        X_test = generate_bert(
            X_test, y_test, text_columns, model, label_encoder)

    return X_train, X_test


def encode_inference(X, y, numerical_columns: list, categorical_columns: list, text_columns: list, prefix: str,
                     text_representation: str, section: str = None) -> DataFrame:
    X = X.reset_index(drop=True)

    if 'empenho_numero_do_processo' in categorical_columns:
        X = generate_process_count(X)
        categorical_columns.remove('empenho_numero_do_processo')
        numerical_columns.append('empenhos_por_processo')

    X = generate_scaler(X, numerical_columns, prefix)

    X = generate_ohe(X, categorical_columns, prefix)

    if text_representation == 'tfidf':
        X = generate_tfidf(X, text_columns, prefix)
    elif text_representation == 'bert':
        label_encoder = load_encoder(f'le_bert_{section}.pkl')
        model = get_saved_model(len(label_encoder.classes_), section)
        X = generate_bert(X, y, text_columns, model, label_encoder)

    return X



import time
import warnings
from datetime import datetime
from functools import partial

from numpy import array
from pandas.core.frame import DataFrame
from tcegoframework.cfgparsing import get_algorithm, get_inference_dataset_path
from tcegoframework.data.filter import scope_filter
from tcegoframework.dremio import construct_query, execute_query
from tcegoframework.io import (load_csv_data, load_model, load_scope_dict,
                               save_inference_plot, save_inference_results)
from tcegoframework.preprocessing.classification import (
    preprocessing_inference_corretude, preprocessing_inference_natureza)
from tcegoframework.preprocessing.text import regularize_columns_name

warnings.filterwarnings('ignore')


def query_dataset(filters: dict) -> DataFrame:
    query = construct_query(filters)
    data = execute_query(query)
    return data


def create_inference_dict(data: DataFrame) -> dict:
    inference_dict = {}
    for i in range(0, data.shape[0]):
        empenho = data.iloc[i, :]
        inference_dict[empenho['empenho_sequencial_empenho']] = {
            'Identificador': empenho['empenho_sequencial_empenho'],
            'Natureza Real': empenho['natureza_despesa_cod'],
            'Natureza Predita': None,
            'Corretude': None,
            'Data Predicao': datetime.now().strftime('%d/%m/%Y'),
            'Resultado': None,
        }
    return inference_dict


def change_inference_dict(out_of_scope: DataFrame, scope_dict: dict, inference_dict: dict) -> dict:
    for _, empenho in out_of_scope.iterrows():
        if empenho['natureza_despesa_cod'] not in scope_dict:
            key = empenho['empenho_sequencial_empenho']
            info = 'Classe desconhecida'
            inference_dict[key]['Natureza Predita'] = info
            inference_dict[key]['Corretude'] = info
            inference_dict[key]['Resultado'] = 'UNK'
        elif empenho['natureza_despesa_cod'] in scope_dict:
            key = empenho['natureza_despesa_cod']
            if scope_dict[key] != 'Em escopo':
                info = scope_dict[key]
                key = empenho['empenho_sequencial_empenho']
                inference_dict[key]['Natureza Predita'] = info
                inference_dict[key]['Corretude'] = info
                info = ''.join([s[0] for s in info.split()]).upper()
                inference_dict[key]['Resultado'] = info
            elif empenho['valor_saldo_do_empenho'] == 0:
                key = empenho['empenho_sequencial_empenho']
                info = 'Saldo zerado'
                inference_dict[key]['Natureza Predita'] = info
                inference_dict[key]['Corretude'] = info
                inference_dict[key]['Resultado'] = 'SNULO'
    return inference_dict


def inference_svm_natureza(data: DataFrame) -> list:
    X = preprocessing_inference_natureza(data.copy(), 'tfidf', 'above')
    model = load_model(filename='svm_natureza_above_model.pkl')
    y_proba_above = model.predict_proba(X)
    y_pred_above = model.predict(X)

    X = preprocessing_inference_natureza(data.copy(), 'tfidf', 'below')
    model = load_model(filename='svm_natureza_below_model.pkl')
    y_proba_below = model.predict_proba(X)
    y_pred_below = model.predict(X)

    y_pred = [
        a if probA >= probB else b
        for a, b, probA, probB in
        zip(y_pred_above, y_pred_below, y_proba_above, y_proba_below)
    ]
    return array(y_pred)


def inference_rf_natureza(data: DataFrame) -> list:
    X = preprocessing_inference_natureza(data.copy(), 'tfidf', 'above')
    model = load_model(filename='rf_natureza_above_model.pkl')
    y_proba_above = model.predict_proba(X)
    y_pred_above = model.predict(X)

    X = preprocessing_inference_natureza(data.copy(), 'tfidf', 'below')
    model = load_model(filename='rf_natureza_below_model.pkl')
    y_proba_below = model.predict_proba(X)
    y_pred_below = model.predict(X)

    y_pred = [
        a if probA.max() >= probB.max() else b
        for a, b, probA, probB in
        zip(y_pred_above, y_pred_below, y_proba_above, y_proba_below)
    ]
    return array(y_pred)


def inference_bert_rf_natureza(data: DataFrame) -> array:
    X = preprocessing_inference_natureza(data.copy(), 'bert', 'above')
    model = load_model(filename='bert_rf_natureza_above_model.pkl')
    y_proba_above = model.predict_proba(X)
    y_pred_above = model.predict(X)

    X = preprocessing_inference_natureza(data.copy(), 'bert', 'below')
    model = load_model(filename='bert_rf_natureza_below_model.pkl')
    y_proba_below = model.predict_proba(X)
    y_pred_below = model.predict(X)

    y_pred = [
        a if probA.max() >= probB.max() else b
        for a, b, probA, probB in
        zip(y_pred_above, y_pred_below, y_proba_above, y_proba_below)
    ]
    return array(y_pred)


def inference_corretude(data: DataFrame) -> array:
    X = preprocessing_inference_corretude(data.copy())
    model = load_model(filename='rf_corretude_model.pkl')
    y_pred_corretude = model.predict(X)
    return y_pred_corretude


def compute_result_code(y_true, y_pred, y_pred_correctness) -> str:
    if y_true == y_pred:
        if y_pred_correctness == 'INCORRETO':
            return 'INCV_M2'
        elif y_pred_correctness == 'OK':
            return 'C_M1-M2'
        elif y_pred_correctness == 'INCONCLUSIVO':
            return 'AD_M2'
    else:
        if y_pred_correctness == 'INCORRETO':
            return 'INCT_M1-M2'
        elif y_pred_correctness == 'OK':
            return 'INCV_M1'
        elif y_pred_correctness == 'INCONCLUSIVO':
            return 'INCV_M1-AD_M2'


def compute_output(data: DataFrame, inference_dict: dict, y_natureza: array, y_corretude: array) -> dict:
    for i in range(data.shape[0]):
        key = data.iloc[i, :]['empenho_sequencial_empenho']
        inference_dict[key]['Natureza Predita'] = y_natureza[i]
        inference_dict[key]['Corretude'] = y_corretude[i]
        inference_dict[key]['Resultado'] = compute_result_code(
            inference_dict[key]['Natureza Real'],
            inference_dict[key]['Natureza Predita'],
            inference_dict[key]['Corretude']
        )
    return inference_dict


def get_dataset(filters: dict) -> DataFrame:
    if dataset_path := get_inference_dataset_path():
        data = load_csv_data(dataset_path)
    else:
        data = query_dataset(filters)
    return data

# define function to parse daterange, date and orgaos filters into filename


def parse_filters(filters: dict) -> str:
    filename = 'inference' + datetime.today().strftime('%d-%m-%Y')
    if 'daterange' in filters:
        start_date, end_date = filters['daterange']
        start_date = start_date.strftime('%d-%m-%Y')
        end_date = end_date.strftime('%d-%m-%Y')
        filename += f'_DR{start_date}-{end_date}'
    if 'dates' in filters:
        dates = filters['dates']
        dates = [date.strftime('%d-%m-%Y') for date in dates]
        dates = '-'.join(dates)
        filename += f'_D{dates}'
    if 'organs' in filters:
        orgaos = filters['orgaos']
        orgaos = '-'.join(orgaos).replace('/', '')
        filename += f'_O{orgaos}'
    return filename


def inference_flow(filters: dict):
    # Executing query
    print('Preparando e executando consulta...')
    data = get_dataset(filters)
    data = data.reset_index(drop=True)
    data = regularize_columns_name(data)

    print('Filtrando e preparando escopo...')
    inference_dict = create_inference_dict(data)
    scope_dict = load_scope_dict('scope.pkl')
    data, out_of_scope = scope_filter(data, scope_dict)

    inference_dict = change_inference_dict(
        out_of_scope, scope_dict, inference_dict)

    if get_algorithm() == 'svm':
        inference_natureza = partial(inference_svm_natureza)
    elif get_algorithm() == 'rf':
        inference_natureza = partial(inference_rf_natureza)
    elif get_algorithm() == 'bert_rf':
        inference_natureza = partial(inference_bert_rf_natureza)

    print('Inferência de Natureza...')
    time_ref = time.time()
    y_pred_natureza = inference_natureza(data.copy())
    print(f'Duração total: {(time.time() - time_ref)/60}')

    print('Inferência de Corretude...')
    time_ref = time.time()
    y_pred_corretude = inference_corretude(data.copy())
    print(f'Duração total: {(time.time() - time_ref)/60}')

    print('Computando e gravando resultados...')
    inference_dict = compute_output(
        data, inference_dict, y_pred_natureza, y_pred_corretude)

    filename = parse_filters(filters) + '.csv'

    results = save_inference_results(filename, inference_dict)
    save_inference_plot(f'{filename}_plot.png', results)
    print('Finalizado.')



# TRIBUNAL DE CONTAS DO ESTADO DE GOIAS
# SECRETARIA DE CONTROLE EXTERNO
# SERVICO DE INFORMACOES ESTRATEGICAS

import time
from subprocess import PIPE, Popen

import pandas as pd
import pexpect
from hdfs3 import HDFileSystem


class TCEHDFS:

    def __init__(self, domain, url, port, user, password):

        try:
            self.domain = domain
            self.url = url
            self.port = port
            self.user = user
            self.pwd = password

            self.popen = Popen(
                ['kdestroy', '-A'],
                stdin=PIPE,
                stdout=PIPE,
                stderr=PIPE)
            self.child = pexpect.spawn(f'kinit {self.user}@{self.domain}')
            self.child.expect('Password*')
            self.child.sendline(str(self.pwd))

            # Esperar 3 segundos para fazer conexao (se nao esperar, emite erro de autenticacao no AD)
            time.sleep(10)
            self.hdfs = HDFileSystem(
                host=self.url,
                port=self.port,
                pars={"hadoop.security.authentication": "kerberos"},
                user=self.user)

        except Exception as e:
            print("Erro ao logar no HDFS. Erro:", e)

    def readHDFS(self, path, file):
        with self.hdfs.open(path + file) as k:
            df = pd.read_csv(k, sep=',', low_memory=False)
            k.close()

        return df

    def writeHDFS(self, path, file, df):

        with self.hdfs.open(path + file, 'wb') as f:
            df.to_csv(f, index=False)
            f.close()

        return path + file

    def copyToHDFS(self, path_local_file, file_name, hdfs_path):
        self.hdfs.put(path_local_file + file_name, hdfs_path + file_name)
        pass
